@@ -83,22 +83,16 @@ err_free:
 static int
 shm_pool_resize(struct shm_pool *pool, int size)
 {
-	if (ftruncate(pool->fd, size) < 0)
+	if (os_resize_anonymous_file(pool->fd, size) < 0)
 		return 0;
 
-#ifdef HAVE_POSIX_FALLOCATE
-	errno = posix_fallocate(pool->fd, 0, size);
-	if (errno != 0)
-		return 0;
-#endif
-
 	wl_shm_pool_resize(pool->pool, size);
 
 	munmap(pool->data, pool->size);
 
 	pool->data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED,
 			  pool->fd, 0);
-	if (pool->data == (void *)-1)
+	if (pool->data == MAP_FAILED)
 		return 0;
 	pool->size = size;
 
@@ -135,7 +129,6 @@ struct wl_cursor_theme {
 	struct wl_cursor **cursors;
 	struct wl_shm *shm;
 	struct shm_pool *pool;
-	char *name;
 	int size;
 };
 
@@ -258,13 +251,10 @@ err_free_cursor:
 }
 
 static void
-load_default_theme(struct wl_cursor_theme *theme)
+load_fallback_theme(struct wl_cursor_theme *theme)
 {
 	uint32_t i;
 
-	free(theme->name);
-	theme->name = strdup("default");
-
 	theme->cursor_count = ARRAY_LENGTH(cursor_metadata);
 	theme->cursors = malloc(theme->cursor_count * sizeof(*theme->cursors));
 
@@ -284,7 +274,7 @@ load_default_theme(struct wl_cursor_theme *theme)
 }
 
 static struct wl_cursor *
-wl_cursor_create_from_xcursor_images(XcursorImages *images,
+wl_cursor_create_from_xcursor_images(struct xcursor_images *images,
 				     struct wl_cursor_theme *theme)
 {
 	struct cursor *cursor;
@@ -345,13 +335,13 @@ wl_cursor_create_from_xcursor_images(XcursorImages *images,
 }
 
 static void
-load_callback(XcursorImages *images, void *data)
+load_callback(struct xcursor_images *images, void *data)
 {
 	struct wl_cursor_theme *theme = data;
 	struct wl_cursor *cursor;
 
 	if (wl_cursor_theme_get_cursor(theme, images->name)) {
-		XcursorImagesDestroy(images);
+		xcursor_images_destroy(images);
 		return;
 	}
 
@@ -371,7 +361,7 @@ load_callback(XcursorImages *images, void *data)
 		}
 	}
 
-	XcursorImagesDestroy(images);
+	xcursor_images_destroy(images);
 }
 
 /** Load a cursor theme to memory shared with the compositor
@@ -397,9 +387,6 @@ wl_cursor_theme_load(const char *name, int size, struct wl_shm *shm)
 	if (!name)
 		name = "default";
 
-	theme->name = strdup(name);
-	if (!theme->name)
-		goto out_error_name;
 	theme->size = size;
 	theme->cursor_count = 0;
 	theme->cursors = NULL;
@@ -411,13 +398,14 @@ wl_cursor_theme_load(const char *name, int size, struct wl_shm *shm)
 	xcursor_load_theme(name, size, load_callback, theme);
 
 	if (theme->cursor_count == 0)
-		load_default_theme(theme);
+		xcursor_load_theme(NULL, size, load_callback, theme);
+
+	if (theme->cursor_count == 0)
+		load_fallback_theme(theme);
 
 	return theme;
 
 out_error_pool:
-	free(theme->name);
-out_error_name:
 	free(theme);
 	return NULL;
 }
@@ -436,7 +424,6 @@ wl_cursor_theme_destroy(struct wl_cursor_theme *theme)
 
 	shm_pool_destroy(theme->pool);
 
-	free(theme->name);
 	free(theme->cursors);
 	free(theme);
 }
@@ -481,7 +468,7 @@ wl_cursor_frame_and_duration(struct wl_cursor *_cursor, uint32_t time,
 	uint32_t t;
 	int i;
 
-	if (cursor->cursor.image_count == 1) {
+	if (cursor->cursor.image_count == 1 || cursor->total_delay == 0) {
 		if (duration)
 			*duration = 0;
 		return 0;
