@@ -261,8 +261,8 @@ wl_global_destroy(struct wl_global *global);
  * A filter function enables the server to decide which globals to
  * advertise to each client.
  *
- * When a wl_global filter is set, the given callback funtion will be
- * called during wl_global advertisment and binding.
+ * When a wl_global filter is set, the given callback function will be
+ * called during wl_global advertisement and binding.
  *
  * This function should return true if the global object should be made
  * visible to the client or false otherwise.
@@ -279,6 +279,12 @@ wl_display_set_global_filter(struct wl_display *display,
 const struct wl_interface *
 wl_global_get_interface(const struct wl_global *global);
 
+uint32_t
+wl_global_get_version(const struct wl_global *global);
+
+struct wl_display *
+wl_global_get_display(const struct wl_global *global);
+
 void *
 wl_global_get_user_data(const struct wl_global *global);
 
@@ -478,6 +484,9 @@ wl_signal_emit(struct wl_signal *signal, void *data)
 		l->notify(l, data);
 }
 
+void
+wl_signal_emit_mutable(struct wl_signal *signal, void *data);
+
 typedef void (*wl_resource_destroy_func_t)(struct wl_resource *resource);
 
 /*
