@@ -25,6 +25,7 @@
 
 #include <stdint.h>
 #include <unistd.h>
+#include <stdatomic.h>
 
 #include "wayland-server.h"
 #include "wayland-client.h"
@@ -46,6 +47,7 @@ struct client_info {
 
 struct display {
 	struct wl_display *wl_display;
+	struct wl_global *test_global;
 
 	struct wl_list clients;
 	uint32_t clients_no;
@@ -64,7 +66,7 @@ struct client {
 	struct wl_display *wl_display;
 	struct test_compositor *tc;
 
-	int display_stopped;
+	atomic_bool display_stopped;
 };
 
 struct client *client_connect(void);
